import math  
import numpy as np  

from keenopt.common.log import funcLog
# from brain.common.config import AlgoConfig
from keenopt.strategy.strategy import Strategy

class Node:
    def __init__(self, index, perturbation):
        super().__init__()
        self.index = index
        self.perturbation = perturbation
        self.retry_count = 0
    

class DFSGradientDescent(Strategy):
    def __init__(self,
                 sample_iteration: int,
                 max_iteration: int,
                 searchspace,
                 fx_weight: np.array,
                 batch_size: int):
        """ Init Grad descent search strategy

        Args:
            fx_weight (np.array):   fx weight vector.
            sample_iteration (int): Sampling iteration.
            max_iteration (int):    Max tuning iteration.
            batch_size (int):       Batch size.
        """
        super().__init__(sample_iteration, max_iteration)

        self.searchspace = searchspace
        self.fx_weight = fx_weight
        self.batch_size = batch_size

        self.min_retry = 1
        self.min_perturbation  = 0.2
        self.grad_inertia_rate = 0.3
        self.init_perturbation = 16
        self.perturbation_reduce_rate = 0.5
        
        self.Stack = []
        
    @funcLog
    def setOrigin(self, 
                   x: np.array, 
                   fx: np.array, 
                   enable_index: list):
        
        """ Get best point as search starting point.

        Use DFS algorithm to traverse best points.
    
        1. Push to self.Stack and replace origin point when we find a configuration with lower loss value, otherwise reduce 
        perturbation by harf.
        
        2. If perturbation reduced to one, retry ‘self.min_retry’ times and backtracking a configuration from self.Stack.
        
        3. Get a new origin point from self.S, if self.Stack is empty.
        
        Args:
            x (np.array):       All configuration to select, shape = n x searchspace.dim.
            fx (np.array):      Value if x, shape = n x fx_dim.
            enable_index(list):  The index of points which failed to generate a valid point.
        """
        
        fx_weighted = np.sum(fx * self.fx_weight,axis=1)
        
        if self.Stack.__len__() == 0:
            sorted_index = np.argsort(np.squeeze(fx_weighted)).tolist()
            
            for index in sorted_index:
                self.Stack.append(
                    Node(index = index, perturbation = self.init_perturbation))
            
            self.batch_node = [self.Stack.pop() for _ in range(self.batch_size)]

        else:
            for i in enable_index:
                # Case 1. Find better point
                if fx_weighted[self.batch_node[i].index] > fx_weighted[len(fx_weighted) - self.batch_size + i]:
                    if self.batch_node[i].perturbation > self.min_perturbation:
                        self.Stack.append(self.batch_node[i])
                    
                    self.batch_node[i] = Node(
                        index = len(fx_weighted) - self.batch_size + i, 
                        perturbation = self.batch_node[i].perturbation
                    )
                
                # Case 2. Fail to find better point
                else:
                    # Case 2.1. Replace seed
                    if self.batch_node[i].perturbation <= self.min_perturbation and self.batch_node[i].retry_count >= self.min_retry:
                        self.batch_node[i] = self.Stack.pop()
                    
                    # Case 2.2 Continue search, retry
                    elif self.batch_node[i].perturbation <= self.min_perturbation:
                        self.batch_node[i].retry_count += 1 
                    
                    # Case 2.3 Continue search, reduce perturbation
                    else:
                        self.batch_node[i].perturbation *= self.perturbation_reduce_rate
                
        best_index = np.array([self.batch_node[i].index for i in range(self.batch_size)])
        return x[best_index,...], fx[best_index,...]
        
        
    @funcLog
    def __perturbPoints(self,
                       surrogate,
                       points: np.array,
                       iteration: int,
                       probability_list:np.array):
        """ Perturbing points and generating a lot of candidate points

        Args:
            surrogate (keenopt.surrogate):  Surrogate model.
            points (np.array):              Perturbing base points. shape = (n, searchspace.dim)
            iteration (int):                Current iteration
            
        Returns:
            np.array: shape = points.shape
        """
        assert points.shape.__len__() == 2
        assert points.shape[1] == self.searchspace.dim
        
        perturbation_list = [self.batch_node[i].perturbation for i in range(self.batch_size)]
        
        # Calculating Gradient.
        grad_matrix = surrogate.calculateGradient(points, self.searchspace, self.fx_weight, scale = perturbation_list)
        
        # Gradient Inertia
        try:
            grad_matrix = self.Grad_Matrix * self.grad_inertia_rate + grad_matrix
        except AttributeError:
            self.Grad_Matrix = grad_matrix
        
        # Perturbing Points
        disturbed_points = self.searchspace.perturbPointsGrad(
            points = points,
            grad_matrix  = grad_matrix,
            perturbation_list = perturbation_list,
            probability_list = probability_list)
        
        return disturbed_points

    @funcLog
    def search(self,
               surrogate,
               x,
               fx,
               iteration: int):
        
        """ Generating configuration points.

        Args:
            surrogate (keenopt.surrogate): Surrogate model.
            x:                             History configure points
            fx:                            Scores of history configure points
            iteration (int):               Current iteration
            
        Return:
            points (list): A list of points 
        """
        assert x.shape[0] > 0
        assert fx.shape[0] > 0
        assert iteration >= self.sample_iteration

        probability_list = np.array([0.7 * (1 - math.log(iteration - self.sample_iteration + 1)/ math.log(self.max_iteration - self.sample_iteration)),] * self.batch_size)
        enable_index = list(range(self.batch_size))
        
        while True:
            origin_points, best_score = self.setOrigin(
                x,
                fx,
                enable_index = enable_index,
            )
            
            enable_index = []
            disturbed_points = self.__perturbPoints(
                surrogate = surrogate,
                points = origin_points,
                iteration = iteration,
                probability_list = probability_list)
            
            dist_matrix = self.searchspace.distance(
                source_points = disturbed_points, 
                target_points = x, 
                concat = 'min')
            
            # for i, dist in enumerate(dist_matrix):
            #     if dist < 0.01:
            #         enable_index.append(i)
            #         probability_list[i] = min(probability_list[i] + 0.1, 1)

            if enable_index.__len__() == 0:
                return disturbed_points
