# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# 
import numpy as np
import os
from keenopt.strategy.strategy import Strategy
from keenopt.common.log import funcLog

class LaMCTS(Strategy):
    @funcLog
    def __init__(self,
                 fx_weight,
                 sample_iteration: int,
                 max_iteration: int,
                 searchspace,
                 batch_size: int,
                 normalize=True,
                 mode='minimize'):
    # def __init__(self,
    #              fx_weight: np.array,
    #              sample_iteration: int,
    #              max_iteration: int,
    #              batch_size: int):

        super().__init__(sample_iteration, max_iteration, mode)
        self.searchspace = searchspace
        self.batch_size = batch_size
        self.normalize = normalize
        self.solver = "bo"

        self.lb = []
        self.ub = []
        for param_info in self.searchspace.param_list:
            if self.normalize:
                lb = 0.0
                ub = 1.0
            else:
                # if param_info.__contains__('options'):
                if hasattr(param_info, "options"):
                    lb = 0
                    ub = param_info.options.__len__() - 1
                # if hasattr(param_info, "range"):
                else:
                    lb = param_info.lb / 1e4
                    ub = param_info.ub / 1e4
            self.lb.append(lb)
            self.ub.append(ub)

        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)

        self.dims = self.searchspace.dim

    @funcLog
    def setOrigin(self,
                  x: np.array,
                  fx: np.array):
        samples = []

        for i in range(x.shape[0]):
            samples.append((x[i], fx[i]*-1 if self.mode=='minimize' else fx[i]))
        return samples

    @funcLog
    def search(self,
               surrogate,
               x: np.array,
               fx: np.array,
               iteration: int):

        """ search configuration points.

        Args:
            surrogate (keenopt.surrogate): Surrogate model.
            iteration (int):               Current iteration

        Return:
            points (list): A list of points
        """
        assert x.shape[0] > 0
        assert fx.shape[0] > 0
        assert iteration >= self.sample_iteration

        while True:
            samples = self.setOrigin(
                x=x,
                fx=fx)

            path = surrogate.select()

            if self.solver == 'turbo':
                samples_flip = []
                for s in samples:
                    samples_flip.append((s[0],s[1]*-1))

                # samples, values = leaf.propose_samples_turbo( 10000, path, self.func )
                proposed_points = surrogate.new_leaf.propose_samples_turbo(
                    self.batch_size,
                    path,
                    self.lb,
                    self.ub,
                    samples_flip)
            else:  # use bo by default
                proposed_points = surrogate.new_leaf.propose_samples_bo(
                    self.batch_size,
                    path,
                    self.lb,
                    self.ub,
                    samples)

            #for i in range(len(proposed_points)):
            #    if self.normalize:
            #        proposed_points[i] = from_unit_cube(self.searchspace, proposed_points[i])

            return proposed_points



