import numpy as np  
import math

from brain.common.config import AlgoConfig

from keenopt.common.log import funcLog
from keenopt.common.tools import normalization
from keenopt.strategy.strategy import Strategy
class Dycors(Strategy):
    def __init__(self,
                 sample_iteration: int,
                 max_iteration: int,
                 searchspace,
                 fx_weight: np.array,
                 batch_size: int,
                 mode: str = 'minimize'):
                
        super().__init__(sample_iteration, max_iteration, mode)
        self.fx_weight = fx_weight
        self.batch_size = batch_size
        self.searchspace = searchspace
        
        self.num_pts = AlgoConfig.keenopt_dycors_num_pts
        self.perturbation_sigma = AlgoConfig.keenopt_dycors_sigma
        
        self.exploration_weight_list = [0.7,0.5,0.2,0.05]
        self.exploration_weight_index = 0
        self.best_index = -1
        self.find_new_best_count = 0

    @funcLog
    def setOrigin(self, 
                  x: np.array, 
                  fx: np.array):
        
        """ Get best point as search starting point.

        If weighting is True, return a point which has min fx.
        If weighting is False, return points with min fx for each fx-dim.
        
        Args:
            x (np.array):         All configuration to select, shape = n x searchspace.dim.
            fx (np.array):        Value if x, shape = n x fx_dim.
        """
        
        min_len = min(x.shape[0],fx.shape[0])
        x  = x[:min_len,...]
        fx = fx[:min_len,...]    
        _, unique_index = np.unique(x,return_index=True,axis=0)
        x = x[unique_index,...]
        fx = fx[unique_index,...]
        
        fx_weighted = np.sum(fx * self.fx_weight,axis=1)
        
        sorted_index = np.argsort(np.squeeze(fx_weighted))
        
        if self.best_index == sorted_index[0]:
            # Not find new best point
            if self.find_new_best_count < 0:
                self.find_new_best_count -= 1
            else:
                self.find_new_best_count = -1
                
        else:
            # Find new best point
            self.best_index = sorted_index[0]
            if self.find_new_best_count > 0:
                self.find_new_best_count += 1
            else:
                self.find_new_best_count = 1
                
        best_point = x[sorted_index[:self.batch_size],...]
        best_score = fx[sorted_index[:self.batch_size],...]
        return best_point, best_score
        

    @funcLog
    def _adjustPerturbationSigma(self, x_dim):
        """ Adjust perturbation sigma value.

        1. Three consecutive rounds of updating the optimal parameters:
        
                sigma^2 = sigma^2 * 2
                
        2. MAX(5, dim) consecutive rounds of not updating the optimal parameters:
        
                sigma^2 = sigma^2 / 2
        """
        if self.find_new_best_count >= 3:
            self.perturbation_sigma *= 2
            self.find_new_best_count = 0
        
        if self.find_new_best_count <= - max(5,x_dim):
            self.perturbation_sigma /= 2
            self.find_new_best_count = 0


    @funcLog
    def _perturbPoints(self,
                       surrogate,
                       points: np.array,
                       iteration: int):
        
        """ Perturbing points and generating a lot of candidate points

        Args:
            surrogate (keenopt.surrogate):  Surrogate model.
            points (np.array):        Perturbing base points. shape = n x searchspace.dim
            iteration (int):                Current iteration
            
        Returns:
            np.array: shape = num_pts x searchspace.dim
        """
        
        disturb_probabilty = min(20 / surrogate.x_dim, 1) * \
            (1 - math.log(iteration - self.sample_iteration + 1) / math.log(self.max_iteration - iteration))
        
        disturbed_points = self.searchspace.perturbPoints(
            num_pts = self.num_pts,
            points = points,
            probabilty   = disturb_probabilty,
            perturbation_sigma = self.perturbation_sigma)

        return disturbed_points


    @funcLog
    def _getExplorationWeight(self):
        """ Get exploration weight
        
        Return the values in the list [0.7,0.5,0.2,0.05] in loop.
        """
        self.exploration_weight_index += 1
        return self.exploration_weight_list[self.exploration_weight_index%len(self.exploration_weight_list)]
    
    
    @funcLog
    def _mostValuablePoints(self,
                            surrogate,
                            x: np.array,
                            points: np.array):
        
        """ Get most valuable points to run benchmark.
        Args:
            surrogate (keenopt.surrogate): Surrogate model.
            x (np.array):                  All configuration which has been evaluated.
            points (np.array):             Configuration points to select.
        """
        # N = points.shape[0]
        # shape = N,
        exploration_weight = self._getExplorationWeight()
        
        dist_matrix = self.searchspace.distance(
            source_points = points, target_points = x, concat = 'min')

        # shape = N * fx_dim
        pred_fx = surrogate.eval(points)
        
        # shape = N,
        weighted_pred_fx = np.mean(pred_fx * self.fx_weight,axis=1)
        
        weighted_value = exploration_weight * normalization(dist_matrix) + \
                        (1 - exploration_weight) * normalization(weighted_pred_fx)
        
        top_index = np.argsort(weighted_value)[::-1][:self.batch_size]
        
        return points[top_index,...], weighted_pred_fx[top_index,...]

        
    @funcLog
    def search(self,
               surrogate,
               x,
               fx,
               iteration: int):
        
        """ Generating configuration points.

        Args:
            surrogate (keenopt.surrogate): Surrogate model.
            iteration (int):               Current iteration.
            
        Return:
            points (list): A list of points 
        """
        assert x.shape[0] > 0
        assert fx.shape[0] > 0
        assert iteration > self.sample_iteration
        
        best_points, _ = self.setOrigin(
            x,
            fx,
        )
        
        self._adjustPerturbationSigma(self.searchspace.dim)
        
        disturbed_points = self._perturbPoints(
            surrogate = surrogate,
            points = best_points,
            iteration = iteration)
        
        if disturbed_points.shape[0] <= self.batch_size:
            return disturbed_points
        
        else:
            points, pred_fx = self._mostValuablePoints(
                x = x,
                points = disturbed_points,
                surrogate = surrogate)
            
            return points