import math
import numpy as np
from abc import ABCMeta,abstractmethod

from keenopt.common.log import funcLog

class Strategy(metaclass=ABCMeta):
    @funcLog
    def __init__(self,
                 sample_iteration: int,
                 max_iteration: int,
                 mode: str = 'minimize',
                 ):

        """ Init Projected Momentum Search (PMS) strategy

        Args:
            fx_weight (np.array):   fx weight vector.
            sample_iteration (int): Sampling iteration.
            max_iteration (int):    Max tuning iteration.
            mode (str):             The optimization direction
            batch_size (int):       Batch size.
        """
        assert mode in ['minimize', 'maximize']
        self.sample_iteration = sample_iteration
        self.max_iteration = max_iteration
        self.mode = mode

    @abstractmethod
    def setOrigin(self,
                  x: np.array,
                  fx: np.array):

        """ Get search starting point.
        Args:
            x (np.array):       All configuration to select, shape = n x searchspace.dim.
            fx (np.array):      Value if x, shape = n x fx_dim.
        Return:
            points (list): A list of points
        """
        ...

    @abstractmethod
    def search(self,
               surrogate,
               x: np.array,
               fx: np.array,
               iteration: int):

        """ search configuration points.

        Args:
            surrogate (keenopt.surrogate): Surrogate model.
            x:                             History configure points
            fx:                            Scores of history configure points
            iteration (int):               Current iteration

        Return:
            points (list): A list of points
        """
        raise NotImplementedError