import numpy as np

from keenopt.common.log import funcLog
from keenopt.surrogate.surrogate import Surrogate
from keenopt.surrogate.models.MLP import MLPregression
from keenopt.surrogate.train.train import train
from keenopt.surrogate.models.MCTSNode import Node
from keenopt.sample_func.sample_func import from_unit_cube

class MCTS(Surrogate):
    @funcLog
    def __init__(self,
                 x_dim: int,
                 fx_dim: int,
                 mode: str = 'minimize'):
        
        super().__init__(x_dim, fx_dim)
        self.mode = mode
        self.x = np.zeros(shape=(0, self.x_dim))
        self.fx = np.zeros(shape=(0, self.fx_dim))

        # self.problem = problem

        # if self.problem.__name__ == "levy":
            # self.Cp = 10
            # self.leaf_size = 8
            # self.kernel_type = "poly"
        # else:
        self.Cp = 1
        self.leaf_size = 10
        self.kernel_type = "rbf"
        self.gamma_type = "scale"
        self.LEAF_SAMPLE_SIZE = self.leaf_size
        self.nodes = []

        self.visualization = False

        self._initModel()

        self.new_leaf = None

    @funcLog
    def _initModel(self):
        # we start the most basic form of the tree, 3 nodes and height = 1

        root = Node(parent=None,
                    dims=self.x_dim,
                    reset_id=True,
                    kernel_type=self.kernel_type,
                    gamma_type=self.gamma_type)
        self.nodes.append(root)
        self.ROOT = root
        self.CURT = self.ROOT

    def eval(self, x: np.array):
        """ Evaluate a point, return prediction value.

        Args:
            x (np.array): A configuration to predict.
        """
        try:
            fx = self.ROOT.classifier.kmean.predict(x)
            if isinstance(fx, float):
                fx = np.array([fx])
            if fx.shape.__len__() == 1:
                fx = np.expand_dims(fx, axis=0)
        except:
            fx = np.zeros((1,1))
        return fx

    def populate_training_data(self,samples):
        # only keep root
        self.ROOT.obj_counter = 0
        for node in self.nodes:
            node.clear_data()
        self.nodes.clear()
        new_root = Node(parent=None,
                        dims=self.x_dim,
                        reset_id=True,
                        kernel_type=self.kernel_type,
                        gamma_type=self.gamma_type)
        self.nodes.append(new_root)

        self.ROOT = new_root
        self.CURT = self.ROOT
        self.ROOT.update_bag(samples)

    def get_leaf_status(self):
        status = []
        for node in self.nodes:
            if node.is_leaf() == True and len(node.bag) > self.LEAF_SAMPLE_SIZE and node.is_svm_splittable == True:
                status.append(True)
            else:
                status.append(False)
        return np.array(status)

    def get_split_idx(self):
        split_by_samples = np.argwhere(self.get_leaf_status() == True).reshape(-1)
        return split_by_samples

    def is_splitable(self):
        status = self.get_leaf_status()
        if True in status:
            return True
        else:
            return False

    def print_tree(self):
        print('-' * 100)
        for node in self.nodes:
            print(node)
        print('-' * 100)

    def reset_to_root(self):
        self.CURT = self.ROOT

    def greedy_select(self):
        self.reset_to_root()
        curt_node = self.ROOT
        path      = [ ]
        if self.visualization == True:
            curt_node.plot_samples_and_boundary(self.problem)
        while curt_node.is_leaf() == False:
            UCT = []
            for i in curt_node.kids:
                UCT.append( i.get_xbar() )
            choice = np.random.choice(np.argwhere(UCT == np.amax(UCT)).reshape(-1), 1)[0]
            path.append( (curt_node, choice) )
            curt_node = curt_node.kids[choice]
            if curt_node.is_leaf() == False and self.visualization == True:
                curt_node.plot_samples_and_boundary(self.problem)
            print("=>", curt_node.get_name(), end=' ' )
        print("")
        return curt_node, path

    def select(self):
        self.reset_to_root()
        curt_node = self.ROOT
        path = []

        while curt_node.is_leaf() == False:
            UCT = []
            for i in curt_node.kids:
                UCT.append(i.get_uct(self.Cp))
            choice = np.random.choice(np.argwhere(UCT == np.amax(UCT)).reshape(-1), 1)[0]
            path.append((curt_node, choice))
            curt_node = curt_node.kids[choice]
            print("=>", curt_node.get_name(), end=' ')
        print("")
        self.new_leaf = curt_node
        return path

    def backpropogate(self, leaf, acc):
        curt_node = leaf
        while curt_node is not None:
            assert curt_node.n > 0
            curt_node.x_bar = (curt_node.x_bar * curt_node.n + acc) / (curt_node.n + 1)
            curt_node.n += 1
            curt_node = curt_node.parent

    @funcLog
    def addPoints(self, x, fx):
        """ Add training data.

        Args:
            x (np.array): configuration numpy matrix.
                            shape = (n, searchspace.dim) or (searchspace.dim, )

            y (np.array): score of the configuration.
                            shape = (n, fx_dim) or (fx_dim, )

        """
        if fx.shape.__len__() == 1:
            fx = np.expand_dims(fx, 0)

        if x.shape.__len__() == 1:
            x = np.expand_dims(x, 0)

        assert x.shape.__len__() == fx.shape.__len__()
        assert x.shape[0] == fx.shape[0]

        if self.new_leaf is not None:
            if self.mode == 'minimize':
                self.backpropogate(self.new_leaf, fx * -1)
            else:
                self.backpropogate(self.new_leaf, fx)

        self.x = np.concatenate((self.x, x), axis=0)
        self.fx = np.concatenate((self.fx, fx), axis=0)



    @funcLog
    def _fit(self,
             x_train,
             fx_train,
             epoch,
             decay,
             reset_model,
             pre_train):
        # we bifurcate a node once it contains over 20 samples
        # the node will bifurcate into a good and a bad kid
        if self.mode == 'minimize':
            _fx_train = fx_train*-1
        else:
            _fx_train = fx_train
        samples = []
        for i in range(x_train.shape[0]):
            samples.append((x_train[i],_fx_train[i]))

        self.populate_training_data(samples)
        assert len(self.ROOT.bag) == len(samples)
        assert len(self.nodes) == 1

        while self.is_splitable():
            to_split = self.get_split_idx()
            # print("==>to split:", to_split, " total:", len(self.nodes) )
            for nidx in to_split:
                parent = self.nodes[nidx]  # parent check if the boundary is splittable by svm
                assert len(parent.bag) >= self.LEAF_SAMPLE_SIZE
                assert parent.is_svm_splittable == True
                # print("spliting node:", parent.get_name(), len(parent.bag))
                good_kid_data, bad_kid_data = parent.train_and_split()
                # creat two kids, assign the data, and push into lists
                # children's lb and ub will be decided by its parent
                assert len(good_kid_data) + len(bad_kid_data) == len(parent.bag)
                assert len(good_kid_data) > 0
                assert len(bad_kid_data) > 0
                good_kid = Node(parent=parent,
                                dims=self.x_dim,
                                reset_id=False,
                                kernel_type=self.kernel_type,
                                gamma_type=self.gamma_type)
                bad_kid = Node(parent=parent,
                               dims=self.x_dim,
                               reset_id=False,
                               kernel_type=self.kernel_type,
                               gamma_type=self.gamma_type)
                good_kid.update_bag(good_kid_data)
                bad_kid.update_bag(bad_kid_data)

                parent.update_kids(good_kid=good_kid, bad_kid=bad_kid)

                self.nodes.append(good_kid)
                self.nodes.append(bad_kid)

            # print("continue split:", self.is_splitable())

        #self.print_tree()