import torch
import torch.nn as nn
import numpy as np

from keenopt.common.log import funcLog
from keenopt.surrogate.surrogate import Surrogate

from keenopt.surrogate.models.MLP import MLPregression
from keenopt.surrogate.train.train import train


class MLPRegressSurrogate(Surrogate):
    def __init__(self, x_dim, fx_dim):
        super().__init__(x_dim, fx_dim)
        self._initModel()

    @funcLog
    def _initModel(self):
        self.model = MLPregression(self.x_dim,self.fx_dim)
        
    def eval(self,x: np.array):
        """ Evaluating configuration point.

        Args:
            x (np.array): configuration point, shape = n x searchspace.dim

        Returns:
            np.array: Evaluating result, shape = n x fx_dim
        """
        if x.shape.__len__() == 1:
            x = np.expand_dims(x,0)
        
        assert len(x.shape) == 2
        assert x.shape[1] == self.x_dim
        
        x = torch.from_numpy(x).cpu().float()
        fx = self.model(x).detach().numpy()
        return fx


    @funcLog
    def _fit(self, 
             x_train: np.array,
             fx_train: np.array,
             epoch = 1,
             decay=0,
             pre_train = False,
             reset_model = False):
        
        """ Training surrogate model.

        If pre_train is True, split data to training data(70%) and eval data(30%) to pre train model and find best epoch. Then train model by full data.
        
        If reset_model is True, reset model to initial model before trainning.

        Args:
            x_train (np.array):    Train data x, shape = n x searchspace.dim
            fx_train (np.array):   Train data fx, shape = n x fx_dim
            epoch (int, optional): Training epoch number. Defaults to 1.
            decay (int, optional): if only use the most recent #decay data to train
            
            pre_train (bool, optional):   If need to pre-training process. Defaults to False.
            reset_model (bool, optional): If need to reset model. Defaults to False.
        """
        
        if reset_model:
            self._initModel()
            self._mean = np.mean(fx_train,axis=0)
            self._std = np.std(fx_train,axis=0)
            
        train(
            x = x_train, 
            fx = fx_train,
            epoch = epoch,
            model = self.model,
            loss_fn = nn.MSELoss(reduction='mean'),
            pre_train = pre_train,
            decay=decay
        )