import numpy as np

from keenopt.common.log import funcLog
from keenopt.surrogate.surrogate import Surrogate

class RBFSurrogate(Surrogate):

    def __init__(self, x_dim, fx_dim):
        self.x_dim = x_dim
        self.fx_dim = fx_dim
        self._initModel()

    @funcLog
    def _initModel(self):
        self.sigma = 1
        self.gaussian_kernel = lambda xi,xj: np.exp(-(xi-xj)**2/2*self.sigma)
        self.matrix = np.zeros(shape = (0,0))
        
        
    def eval(self,x: np.array):
        """ Evaluating configuration point.

        Args:
            x (np.array): configuration point, shape = n x searchspace.dim

        Returns:
            np.array: Evaluating result, shape = n x fx_dim
        """
        if x.shape.__len__() == 1:
            x = np.expand_dims(x,0)
                
        y_pred = np.zeros(shape = (x.shape[0],self.fx_dim))

        try:
            if self.w.shape[0] != self.x.shape[0]:
                self._fitImpl()
            
            for i in range(x.shape[0]):
                for k in range(self.x.shape[0]):
                    v = self.w[k] * np.mean(self.gaussian_kernel(x[i,...],self.x[k,...]))
                    v = np.squeeze(np.array(v))
                    y_pred[i] += v
                    
            return y_pred
                    
        except AttributeError:
            return np.zeros(shape=(x.shape[0],self.fx_dim))
        

    def _fitImpl(self):
        """ Training RBF kernel matrix.
        
        General process if RBF Interpolation function: 
        1. Calculating matrix[i][j] by gaussian kernel function : f(xi,xj) = exp(-(xi-xj)**2/2*sigma)
        2. Calculating W matrix by W = matrix^-1 * Y
        
        Time complexity of RBF Interpolation function fitting is upto O(N^3), because we need to recalculate relationship matrix in each iteration.
        
        We realized a increment-calculating method to reduce Time complexity to O(N^2)
        """
        dim = self.matrix.shape[0]
        delta = self.x.shape[0] - dim
        expand_matrix_right = np.zeros(shape = (dim,delta))
        for i in range(dim):
            for k in range(delta):
                expand_matrix_right[i][k] = np.mean(self.gaussian_kernel(self.x[i,...],self.x[dim + k,...]))
        
        expand_matrix_down = np.zeros(shape = (delta, dim + delta))
        for i in range(delta):
            for k in range(dim + delta):
                expand_matrix_down[i][k] = np.mean(self.gaussian_kernel(self.x[dim+i,...],self.x[k,...]))
                
        if dim == 0:
            self.matrix = expand_matrix_down
            
        else:
            self.matrix = np.concatenate(
                (self.matrix, expand_matrix_right), axis=1)
            self.matrix = np.concatenate(
                (self.matrix, expand_matrix_down),  axis=0)        
        
        _, unique_index = np.unique(self.x,return_index=True,axis=0)
        
        self.x = self.x[unique_index,...]
        self.fx = self.fx[unique_index,...]
        self.matrix = self.matrix[unique_index,...][...,unique_index]
        self.w = np.asmatrix(self.matrix).I * np.asmatrix(self.fx)
    
    
    @funcLog
    def _fit(self, 
             x_train: np.array,
             fx_train: np.array,
             epoch = 1, 
             pre_train = False,
             reset_model = False):
        self._fitImpl()