import time
from copy import deepcopy
from keenopt.common.log import funcLog

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as Data
import torch.nn.functional as F

@funcLog
def _trainImpl(model,
              x_train: np.array,
              y_train: np.array,
              x_eval: np.array,
              y_eval: np.array,
              batch_size: int,
              epoch: int,
              loss_fn: torch.nn.Module,
              optimizer: torch.optim):
    
    """ Training a model and eval best epoch.

    Args:
        model (keenopt.surrogate.model): Surrogate model.
        x_train (np.array): Training data x.
        y_train (np.array): Training data fx.
        x_eval (np.array):  Evaluation data x.
        y_eval (np.array):  Evaluation data fx.
        batch_size (int):   Batch size.
        epoch (int):        Epoch
        loss_fn (torch.nn.Module): Loss function.
        optimizer (torch.optim):   Optimizer.

    Returns:
        int: Best epoch.
    """

    data_loader = Data.DataLoader(
        dataset = Data.TensorDataset(x_train, y_train),
        batch_size = batch_size,
        shuffle = True,
        # num_workers = 2,
    )
    
    model.eval()
    eval_loss = np.zeros(shape=(epoch,))
    worse_count = 0
    
    for i in range(epoch):
        model.train()
        s_tic = time.time()
        for batch_x, batch_y in data_loader:
            d_tac = time.time()

            optimizer.zero_grad()
            pred_y = model(batch_x)
            loss = loss_fn(pred_y, batch_y)
            f_tac = time.time()
            #optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            # print(f'data_load time: {d_tac - s_tic}, forward time: {f_tac - d_tac}, backward time: {time.time() - f_tac}')
            s_tic = time.time()

            
        model.eval()
        eval_loss[i] = loss_fn(model(x_eval), y_eval)
        # TODO: adjust worse_count and rate can control training accuracy
        if eval_loss[i] > np.min(eval_loss[:i+1]) * 1.1:
            worse_count += 1
        else:
            worse_count = 0
            
        if worse_count > 5:
            break
    
    return np.argmin(eval_loss)

@funcLog
def train(model, 
          x, 
          fx,
          loss_fn,
          epoch,
          lr = 1e-3,#0.005
          pre_train = False,
          batch_size = 16,
          eval_data_rate = 0.3,
          decay=0
          ):
    """ Training model.

    Args:
        model (keenopt.surrogate.model): Surrogate model.
        x (np.array):  Training data x.
        fx (np.array): Training data fx.
        loss_fn (torch.nn.Module): Loss function.
        epoch (int): Epoch
        lr (float, optional): learning rate. Defaults to 0.005.
        pre_train (bool, optional): if run pre-training. Defaults to False.
        batch_size (int, optional): Batch size. Defaults to 16.
        eval_data_rate (float, optional): Rate of eval dataset. Defaults to 0.3.
    """ 
    #device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    device = torch.device("cpu")
    
    if pre_train:
        eval_model = deepcopy(model)
        eval_optimizer = optim.Adam(eval_model.parameters(), lr = lr)
        
        eval_num = int(x.shape[0] * eval_data_rate)
        x_train = torch.from_numpy(x[:x.shape[0] - eval_num,...]).cpu().float()
        y_train = torch.from_numpy(fx[:x.shape[0] - eval_num,...]).cpu().float()
        x_eval = torch.from_numpy(x[-eval_num:,...]).cpu().float()
        y_eval = torch.from_numpy(fx[-eval_num:,...]).cpu().float()
        
        best_epoch = _trainImpl(
            model = eval_model,
            x_train = x_train,
            y_train = y_train,
            x_eval  = x_eval,
            y_eval  = y_eval,
            epoch   = epoch,
            loss_fn = loss_fn,
            batch_size = batch_size,
            optimizer  = eval_optimizer
        )
    else:
        best_epoch = epoch
    
    x_train = torch.from_numpy(x).to(device).float()
    y_train = torch.from_numpy(fx).to(device).float()
    optimizer = optim.Adam(model.parameters(), lr = lr)

    if decay > 0:
        n_train = x_train.shape[0] if x_train.shape[0] <= decay else decay
        x_train = x_train[-n_train:,...]
        y_train = y_train[-n_train:,...]
    
    _trainImpl(
        model = model,
        x_train = x_train,
        y_train = y_train,
        x_eval  = x_train,
        y_eval  = y_train,
        epoch   = best_epoch,
        loss_fn = loss_fn,
        batch_size = batch_size,
        optimizer  = optimizer
    )